from .NodeStyle import NodeStyle, DEFAULT_FONT, DEFAULT_SHAPE, DEFAULT_SHAPE_STYLE
from colouration import Colour

def generate_diverging_numbers(n):
	if n == 1:
		return [0]
	elif n % 2 == 0:
		return [x/(n/2) for x in range(1, n // 2 + 1)] + [-x/(n/2) for x in range(1, n // 2 + 1)]
	else:
		return [0] + generate_diverging_numbers(n-1)

def get_diverging_number(i, n, reverse):
	return sorted(generate_diverging_numbers(n), reverse=reverse)[i]

def mix_node_style_of_parents(node, font=DEFAULT_FONT, shape=DEFAULT_SHAPE, shape_style=DEFAULT_SHAPE_STYLE):
	"""
	:type node: abstract.Node.Node
	:rtype: NodeStyle
	"""
	scheme = node.graph.colour_scheme
	style = None
	colour = None
	if node.has_parents():
		candidates = []
		for parent in node.parents:
			if not parent.is_in_loop() or not parent.is_style_callable():
				style = parent.compiled_style.copy()
				candidates.append((style, node.get_child_rank(parent=parent), len(parent.children)))

		if len(candidates) > 0:
			colour = Colour.mix(
				colours=[
					style.colour.increase_hue(amount=0.1*get_diverging_number(i=rank, n=num_siblings, reverse=False))
					for style, rank, num_siblings in candidates
				]
			)
			colour = colour.pale(ratio=0.2)
			colour.weight = colour.weight**(0.5)
			#colour = colour.mix_with_gray(gray_weight=2)
			style = NodeStyle(colour=colour, font=font, shape=shape, shape_style=shape_style)

	if style is None or colour is None:
		colour = scheme.least_used_colour
		style = NodeStyle(colour=colour, font=font, shape=shape, shape_style=shape_style)

	return style


from .NodeStyle import NodeStyle
from .EdgeStyle import EdgeStyle

from colouration import Scheme, Colour
DEFAULT_BACKGROUND_COLOUR = '#FAFAFA'
DEFAULT_COLOUR_SCHEME = 'pastel15'


class GraphStyle:
	def __init__(self, graph):
		self._node_colour_indices = None
		self._node_styles = {}
		self._edge_styles = {}
		self._colour_scheme = None
		self._node_styles_counter = {}
		self._edge_styles_counter = {}
		self._background_colour = None
		self._graph = graph
		self.add_colours(
			name='default', colour_scheme=DEFAULT_COLOUR_SCHEME, background_colour=DEFAULT_BACKGROUND_COLOUR
		)

	def __getstate__(self):
		return {
			'node_styles': self._node_styles,
			'edge_styles': self._edge_styles,
			'node_styles_counter': self._node_styles_counter,
			'edge_styles_counter': self._edge_styles_counter,
			'node_colour_indices': self._node_colour_indices,
			'background_colour': self._background_colour
		}

	def __setstate__(self, state):
		self._node_styles = state['node_styles']
		self._edge_styles = state['edge_styles']
		self._node_styles_counter = state['node_styles_counter']
		self._edge_styles_counter = state['edge_styles_counter']
		self._node_colour_indices = state['node_colour_indices']
		self._background_colour = state['background_colour']

	@property
	def graph(self):
		"""
		:rtype: .Graph.Graph
		"""
		return self._graph

	def reset_node_colours(self):
		self._node_colour_indices = None

	def add_colours(self, name, colour_scheme, background_colour):
		colour_scheme = Scheme.auto(obj=colour_scheme or DEFAULT_COLOUR_SCHEME)
		self._background_colour = Colour.auto(obj=background_colour)
		self._node_styles[name] = [
			create_node_colour_inheritance_style(colour=colour)
			for colour in colour_scheme.colours
		]
		self._edge_styles[name] = [
			inherit_colour_from_start_node
			for colour in colour_scheme.colours
		]

	def get_node_style(self, style_name, number=None, node_name=None):
		node_style = self._node_styles[style_name]
		if isinstance(node_style, NodeStyle):
			return node_style
		else:
			if number is None:
				number = self.node_colour_indices[node_name]
			return node_style[number % len(node_style)]

	def get_edge_style(self, style_name, number=None, node_name=None):
		edge_style = self._edge_styles[style_name]
		if isinstance(edge_style, EdgeStyle):
			return edge_style
		else:
			if number is None:
				number = self.node_colour_indices[node_name]
			return edge_style[number % len(edge_style)]

	@property
	def node_colour_indices(self):
		if self._node_colour_indices is None:
			sorted_nodes = [
				name for name, _ in sorted(
					self.graph.nodes_dict.items(),
					key=lambda name_node: (-name_node[1].num_outward_edges, name_node[1].index)
				)
			]
			self._node_colour_indices = {name: index for index, name in enumerate(sorted_nodes)}

		return self._node_colour_indices

	@property
	def node_styles(self):
		"""
		:rtype: dict[NodeStyle]
		"""
		return self._node_styles

	@property
	def edge_styles(self):
		"""
		:rtype: dict[EdgeStyle]
		"""
		return self._edge_styles

	@property
	def background_colour(self):
		"""
		:rtype: Colour
		"""
		return self._background_colour

	def __eq__(self, other):
		"""
		:type other: GraphStyle
		:rtype: bool
		"""
		condition_1 = self.edge_styles == other.edge_styles
		condition_2 = self.node_styles == other.node_styles
		condition_3 = self.background_colour == other.background_colour
		return condition_1 and condition_2 and condition_3

	def __ne__(self, other):
		return not self.__eq__(other=other)



from .StyleMaster import NodeStylist, EdgeStylist, StyleMaster
from .. import Graph

class NodeColourInverter(NodeStylist):
	def __init__(self):
		super().__init__(style=None, condition=None)

	def paint(self, graph):
		"""
		:type graph: Graph
		:rtype: Graph
		"""
		for node in graph.nodes:
			node.style = node.style.invert_colours()

		return graph

	def copy(self):
		return self.__class__()


class EdgeColourInverter(EdgeStylist):
	def __init__(self):
		super().__init__(style=None, condition=None)

	def paint(self, graph):
		"""
		:type graph: Graph
		:rtype: Graph
		"""
		for edge in graph.edges:
			edge.style = edge.style.invert_colours()

		return graph

	def copy(self):
		return self.__class__()


class ColourInverter(StyleMaster):
	def __init__(self):
		super().__init__(stylists=[NodeColourInverter(), EdgeColourInverter()])

	def paint(self, graph):
		"""
		:type graph: Graph
		:rtype: Graph
		"""

		for stylist in self.stylists:
			stylist.paint(graph=graph)
		graph.background_colour = graph.background_colour.invert()
		return graph

	def copy(self):
		return self.__class__()



from .StyleMaster import NodeStylist
from .NodeStyle import NodeStyle
from .. import Node
from colouration import Scheme
DEFAULT_COLOUR_SCHEME = 'pastel15'


def is_root_or_parents_are_in_loop(node):
	"""
	:type node: Node
	:rtype: bool
	"""
	parents = [parent for parent in node.parents if not parent.is_in_loop()]
	return len(parents) == 0


class RootStylist(NodeStylist):
	def __init__(self, node_style=None, colour_scheme=None, strict=False):
		"""
		:type colour_scheme: Scheme or NoneType or str
		"""
		self._strict = strict # strict means number of descendants is also important
		self._node_style = node_style

		if isinstance(colour_scheme, str):
			colour_scheme = Scheme(name=colour_scheme)
		elif colour_scheme is None:
			colour_scheme = Scheme(name=DEFAULT_COLOUR_SCHEME)

		self._colour_scheme = colour_scheme
		self._colours_used = {colour: 0 for colour in self._colour_scheme.colours}

		super().__init__(style=None, condition=is_root_or_parents_are_in_loop)

	def _get_least_used_colour(self, usage=1):
		colour = min(self._colours_used, key=self._colours_used.get)
		self._use_colour(colour=colour, usage=usage)
		return colour

	def _use_colour(self, colour, usage=1):
		self._colours_used[colour] += usage

	def paint(self, graph):
		"""
		:type graph: Graph
		:rtype Graph
		"""

		for node in graph.nodes:
			if self._condition_function(node):
				if self._strict:
					usage = node.num_descendants
				else:
					usage = 1

				colour = self._get_least_used_colour(usage=usage)
				if self._node_style is None:
					style = NodeStyle(colour=colour)
				else:
					style = self._node_style.copy()
					style.reset_colours()
					style.colour = colour

				node.style = style

		return graph

	def copy(self):
		return self.__class__(colour_scheme=self._colour_scheme, strict=self._strict)



from .StyleMaster import NodeStylist, EdgeStylist, StyleMaster
from .NodeStyle import NodeStyle
from .EdgeStyle import EdgeStyle


class LoopNodeStylist(NodeStylist):
	def __init__(self, style=None):
		style = style or NodeStyle(fill_colour='yellow', border_colour='red', colour='red', shape='circle')
		super().__init__(style=style, condition=lambda node: node.is_in_loop())

	def copy(self):
		return self.__class__(style=self._style)


class LoopEdgeStylist(EdgeStylist):
	def __init__(self, style=None):
		style = style or EdgeStyle(colour='red')
		super().__init__(style=style, condition=lambda edge: edge.is_in_loop())

	def copy(self):
		return self.__class__(style=self._style)


class LoopStylist(StyleMaster):
	def __init__(self, node_style=None, edge_style=None):
		self._node_style = node_style
		self._edge_style = edge_style
		super().__init__(stylists=[LoopNodeStylist(style=node_style), LoopEdgeStylist(style=edge_style)])

	def copy(self):
		return self.__class__(node_style=self._node_style, edge_style=self._edge_style)





def _get_ancestors(obj, _objects_travelled=None, max_height=None):
	"""
	:type _objects_travelled: list or NoneType
	:rtype: list[Node]
	"""
	if max_height is not None:
		if max_height <= 0:
			return [], {}
		else:
			max_height -= 1
	_objects_travelled = _objects_travelled or []
	_objects_travelled.append(obj)
	parents = obj.__parents__()
	if len(parents) == 0:
		return [], {}
	else:
		ancestors = []
		ancestors_dict = {1: []}
		for parent in parents:
			if parent not in _objects_travelled:
				if parent not in ancestors:
					ancestors.append(parent)
					ancestors_dict[1].append(parent)
				parent_ancestors, parent_ancestors_dict = _get_ancestors(
					obj=parent, _objects_travelled=_objects_travelled, max_height=max_height
				)
				for ancestor in parent_ancestors:
					if ancestor not in ancestors:
						ancestors.append(ancestor)
						for distance in parent_ancestors_dict.keys():
							ancestors_dict[distance + 1] = parent_ancestors_dict[distance]
		return ancestors, ancestors_dict


def get_ancestors(obj, distance=True, max_height=None):
	ancestors, ancestors_dict = _get_ancestors(obj=obj, _objects_travelled=None, max_height=max_height)
	if distance:
		return ancestors_dict
	else:
		return ancestors
